"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

import operator
import time

from twisted.internet import defer

from smap.util import build_recursive
from smap.contrib import dtutil

import querygen as qg
from data import escape_string
from querydata import extract_data
import data
import stream

import ply
import ply.lex as lex
import ply.yacc as yacc
import copy

tokens = (
    'HAS', 'AND', 'OR', 'NOT', 'LIKE', 'DISTINCT', 'STAR',
    'LPAREN', 'RPAREN', 'COMMA', 'TAGS', 'SELECT', 'WHERE',
    'QSTRING', 'EQ', 'NUMBER', 'LVALUE', 'IN', 'DATA', 'DELETE',
    'SET', 'TILDE', 'BEFORE', 'AFTER', 'NOW', 'LIMIT', 'STREAMLIMIT',
    'APPLY', 'TO', 'AS', 'GROUP', 'BY', 'JOIN', 'ON',
    )

precedence = (
    ('left', 'LIKE', 'TILDE', 'IN'),
    ('left', 'AND'),
    ('left', 'OR'),
    ('right', 'EQ'),
    ('right', 'NOT'),
    ('left', 'COMMA')
    )

reserved = {
    'where' : 'WHERE',
    'distinct' : 'DISTINCT',
    'select' : 'SELECT',
    'delete' : 'DELETE',
    'set' : 'SET',
    'tags' : 'TAGS',
    'has' : 'HAS',
    'and' : 'AND',
    'or' : 'OR',
    'not' : 'NOT',
    'like' : 'LIKE',
    'data': 'DATA',
    'in' : 'IN',
    'before' : 'BEFORE',
    'after' : 'AFTER',
    'now' : 'NOW',
    'limit' : 'LIMIT',
    'streamlimit' : 'STREAMLIMIT',
    '~' : 'TILDE',
    'apply' : 'APPLY',
    'to' : 'TO',
    'as' : 'AS',
    'group' : 'GROUP',
    'join' : 'JOIN',
    'on' : 'ON',
    'by' : 'BY',
    }

t_LPAREN = r'\('
t_RPAREN = r'\)'
t_STAR = r'\*'
t_COMMA = r','
def t_QSTRING(t):
    r'("[^"\\]*?(\\.[^"\\]*?)*?")|(\'[^\'\\]*?(\\.[^\'\\]*?)*?\')'    
    if t.value[0] == '"':
        t.value = t.value[1:-1].replace('\\"', '"')
    elif t.value[0] == "'":
        t.value = t.value[1:-1].replace("\\'", "'")
    return t

t_EQ = r'='

def t_LVALUE(t):
    r'[a-zA-Z\~\$][a-zA-Z0-9\/\%_\-]*'
    t.type = reserved.get(t.value, 'LVALUE')
    return t


def t_NUMBER(t):
    r'([0-9]+)|([+-]?([0-9]*\.)?[0-9]+)'
    if '.' in t.value:
        try:
            t.value = float(t.value)
        except ValueError:
            print "Invalid floating point number", t.value
            t.value = 0
    else:
        try:
            t.value = int(t.value)
        except ValueError:
            print "Integer value too large %d", t.value
            t.value = 0
        
    return t

t_ignore = " \t"
def t_newline(t):
    r'[\n\r]+'
    t.lexer.lineno += t.value.count("\n")

def t_error(t):
    print("Illegal character '%s'" % t.value[0])
    t.lexer.skip(1)

smapql_lex = lex.lex()

# precedence = ('and')
names = {}

def ext_default(x):
    return map(operator.itemgetter(0), x)

def ext_deletor(x):
    data.del_streams(map(operator.itemgetter(0), x))
    return map(operator.itemgetter(1), x)

def ext_plural(x):
    rv = {}
    for uuid, tagname, tagval in x:
        if not uuid in rv:
            rv[uuid] = {'uuid' : uuid}
        rv[uuid][tagname] = tagval
    return map(lambda x: build_recursive(x, suppress=[]), rv.itervalues())

TIMEZONE_PATTERNS = [
    "%m/%d/%Y",
    "%m/%d/%Y %M:%H",
    "%Y-%m-%dT%H:%M:%S",
    ]
def parse_time(ts):
    for pat in TIMEZONE_PATTERNS:
        try:
            dt = dtutil.strptime_tz(ts, pat)
            return dtutil.dt2ts(dt) * 1000
        except ValueError:
            continue
    raise ValueError("Invalid time string:" + ts)

def make_select_rv(t, sqlsel, wherestmt=None):
    if wherestmt != None:
        return sqlsel[2], ("""SELECT %s FROM metadata2 m, stream s 
              WHERE stream_id IN (%s) AND %s AND
                   s.id = m.stream_id""" % \
                               (sqlsel[0], qg.normalize(wherestmt).render(), 
                                sqlsel[1]))
    else:
        return sqlsel[2], ("""SELECT %s FROM metadata2 m, stream s, subscription sub 
              WHERE s.id = m.stream_id AND s.subscription_id = sub.id
                 AND %s AND %s""" % 
                           (sqlsel[0], sqlsel[1], 
                            qg.build_authcheck(t.parser.request)))

def p_query(t):
    '''query : SELECT selector WHERE statement
             | SELECT selector
             | SELECT data_clause WHERE statement
             | DELETE tag_list WHERE statement
             | DELETE WHERE statement
             | SET set_list WHERE statement
             | APPLY apply_statement
             '''
    if t[1] == 'select': 
        if len(t) == 5:
            t[0] = make_select_rv(t, t[2], t[4])
        elif len(t) == 3:
            t[0] = make_select_rv(t, t[2])
        
    elif t[1] == 'delete':
        # a new delete inner statement enforces that we only delete
        # things which we have the key for.
        delete_inner = """
   (SELECT s.id FROM stream s, subscription sub WHERE s.id IN %s
          AND s.subscription_id = sub.id AND %s)
    """ % (qg.normalize(t[len(t) - 1]).render(), 
           qg.build_authcheck(t.parser.request, forceprivate=True))
        if t[2] == 'where':
            # delete the whole stream, gone.  this also deletes the
            # data in the backend readingdb.
            t[0] = ext_deletor, \
                """DELETE FROM stream s WHERE id IN %s
                   RETURNING s.id, s.uuid
                """ % delete_inner
        else:
            # this alters the tags but doesn't touch the data
            t[0] = None, \
                ("""DELETE FROM metadata2 WHERE stream_id IN %s
                    AND (%s)""" % (delete_inner, 
                                   ' OR '.join(map(lambda x: "tagname = %s" % 
                                                   escape_string(x), 
                                                   t[2]))))
    elif t[1] == 'set':
        #  set tags by calling the add_tag stored procedure with each
        #  new tag; this'll insert or update the database as
        #  appropriate
        tag_stmt = ','.join(map(lambda (t, v): \
                                    "add_tag(m.stream_id, %s, %s)" % 
                                (escape_string(t), 
                                 escape_string(v)), 
                                t[2]))
        # filter by the selector; adding an authcheck which only lets
        # you operate on *your* streams.
        t[0] = None, \
            """SELECT %s FROM stream s, metadata2 m, subscription sub
               WHERE stream_id IN %s AND %s AND
                 s.id = m.stream_id AND s.subscription_id = sub.id
            """ % \
            (tag_stmt,
             qg.normalize(t[4]).render(),
             qg.build_authcheck(t.parser.request, forceprivate=True))
    elif t[1] == 'apply':
        t[0] = t[2]

def p_apply_statement(t):
    """apply_statement  : apply_clause TO data_clause WHERE statement_as_list
                        | apply_clause TO data_clause WHERE statement_as_list GROUP BY tag_list
    """
    if len(t[5]) != 1:
        raise Exception("Only one group supported now...")

    tag_extractor, tag_query = make_select_rv(t, make_tag_select('*'), t[5][0][1])
    data_extractor, data_query = make_select_rv(t, t[3], t[5][0][1])

    if len(t) > 7: group = t[8]
    else: group = None
    app = stream.make_applicator(t[1], group=group)

    t[0] = [app, tag_extractor, data_extractor], [None, tag_query, data_query]

def make_tag_select(taglist):
    select = "s.uuid, m.tagname, m.tagval"
    if taglist  == '*':
        restrict = 'true'
    else:
        def make_clause(x):
            if x == 'uuid':
                return "tagname = 'Path'"
            else:
                return "tagname = %s" % escape_string(x)
        restrict = ('(' + " OR ".join(map(make_clause, taglist)) + ')')
    return (select, restrict, ext_plural)

def p_selector(t):
    '''selector : tag_list
                | STAR
                | DISTINCT LVALUE
                | DISTINCT TAGS'''
    if t[1] == 'distinct':
        if t[2] == 'tags':
            restrict = 'true'
            t[0] =("DISTINCT m.tagname", 'true', ext_default)
        elif t[2] == 'uuid':
            t[0] = ("DISTINCT s.uuid", "true", ext_default)
        else:
            t[0] = ("DISTINCT m.tagval", "tagname = %s" % 
                     escape_string(t[2]), ext_default)
    else:
        t[0] = make_tag_select(t[1])

def p_data_clause(t):
    '''data_clause : DATA IN LPAREN timeref COMMA timeref RPAREN limit
                   | DATA BEFORE timeref limit
                   | DATA AFTER timeref limit'''
    if t[2] == 'in':
        method = 'data'
        start, end = t[4], t[6]
        limit = t[8]
        if limit[0] == None: limit[0] = 10000
    elif t[2] == 'before':
        method = 'prev'
        start, end = t[3], 0
        limit = t[4]
        if limit[0] == None: limit[0] = 1
    elif t[2] == 'after':
        method = 'next'
        start, end = t[3], 0
        limit = t[4]
        if limit[0] == None: limit[0] = 1

    t[0] = ("distinct(s.uuid), s.id", "true", 
            lambda streams: extract_data(streams, method, start, end,
                                         limit[0], limit[1]))

def p_timeref(t):
    '''timeref : NUMBER 
               | QSTRING
               | NOW'''
    if t[1] == 'now':
        t[0] = int(time.time()) * 1000
    elif type(t[1]) == type(''):
        t[0] = parse_time(t[1])
    else:
        t[0] = t[1]

def p_limit(t):
    '''limit : 
             | LIMIT NUMBER
             | STREAMLIMIT NUMBER
             | LIMIT NUMBER STREAMLIMIT NUMBER'''
    limit, slimit = [None, 10]
    if len(t) == 1:
        pass
    elif t[1] == 'limit':
        limit = t[2]
        if len(t) == 5:
            slimit = t[4]
    elif t[1] == 'streamlimit':
        slimit = t[2]
    t[0] = [limit, slimit]

def p_apply_clause(t):
    '''apply_clause   : LVALUE LPAREN RPAREN
                      | LVALUE LPAREN call_list RPAREN
                      | LVALUE LPAREN apply_clause RPAREN
                      | LVALUE LPAREN apply_clause COMMA call_list RPAREN
                     '''
    if len(t) == 4:
        t[0] = [stream.get_operator(t[1], [])]
    elif len(t) == 5:
        if len(t[3]) > 0 and t[3][0] == 'args':
            t[0] = [stream.get_operator(t[1], t[3][1])]
        else:
            t[0] = t[3] + [stream.get_operator(t[1], [])]
    elif len(t) == 7:
        t[0] = t[3] + [stream.get_operator(t[1], t[5][1])]

def p_call_list(t):
    '''call_list     : QSTRING
                     | NUMBER
                     | LVALUE
                     | QSTRING COMMA call_list
                     | NUMBER COMMA call_list
                     | LVALUE COMMA call_list
                     '''
    if len(t) == 2:
        t[0] = ('args', [t[1]])
    else:
        t[0] = ('args', [t[1]] + t[3][1])

def p_tag_list(t):
    '''tag_list : LVALUE
                | LVALUE COMMA tag_list'''
                
    if len(t) == 2:
        t[0] = [t[1]]
    else:
        t[0] = [t[1]] + t[3]

def p_set_list(t):
    '''set_list : LVALUE EQ QSTRING
                | LVALUE EQ QSTRING COMMA set_list'''
    if len(t) == 4:
        t[0] = [(t[1], t[3])]
    else:
        t[0] = [(t[1], t[3])] + t[5]

def p_statement_as_list(t):
    """statement_as_list   : statement
                           | statement AS LVALUE
                           | statement_as_list COMMA statement
                           | statement_as_list COMMA statement AS LVALUE 
                           """
    if len(t) == 2:
        t[0] = [('$1', t[1])]
    elif len(t) == 4 and t[2] == 'as':
        t[0] = [(t[3], t[1])]
    elif len(t) == 4 and t[2] == ',':
        t[0] = t[1] + [("$%i" % (len(t[1]) + 1), t[3])]
    else:
        t[0] = t[1] + [(t[5], t[3])]

def merge_clauses(klass, lstmt, rstmt) :
    if type(lstmt) == klass and type(rstmt) == klass:
        return klass(lstmt.clauses.union(rstmt.clauses))
    elif type(lstmt) == klass:
        return klass(lstmt.clauses.union(set([rstmt])))
    elif type(rstmt) == klass:
        return klass(rstmt.clauses.union(set([lstmt])))
    else:
        return klass([lstmt, rstmt])

def p_statement(t):
    '''statement : statement_unary
                 | statement_binary
                 | LPAREN statement RPAREN
                 | statement AND statement
                 | statement OR statement
                 | NOT statement
                '''
    if len(t) == 2:
        t[0] = t[1]
    elif t[2] == 'and':
        t[0] = merge_clauses(qg.AndOperator, t[1], t[3])
    elif t[2] == 'or':
        t[0] = merge_clauses(qg.OrOperator, t[1], t[3])
    elif t[1] == 'not':
        t[0] = qg.NotOperator([t[2]])
    else:
        t[0] = t[2]

def p_statement_unary(t):
    '''statement_unary : HAS LVALUE'''
    if t[1] == 'has':
        t[0] = qg.Clause(qg.build_clause(t.parser.request, 
                                         "tagname = %s" % escape_string(t[2])))

def p_statement_binary(t):
    '''statement_binary : LVALUE EQ QSTRING
                        | LVALUE LIKE QSTRING
                        | LVALUE TILDE QSTRING
                        | LVALUE IN LPAREN statement RPAREN
    '''
    if t[2] == 'in':
        t[0] = qg.Clause(qg.build_clause(t.parser.request,
                                         """
  mi1.tagname = %s AND
  si1.uuid = mi1.tagval AND
  si1.id IN (%s)""" % (t[1], t[4]), ti='1'))
    elif t[1] == 'uuid':
        t[0] = qg.Clause(qg.build_clause(t.parser.request, "(si.uuid = %s)" %
                                         escape_string(t[3])))
    elif t[2] == '=':
        t[0] = qg.Clause(qg.build_clause(t.parser.request, "(tagname = %s AND tagval = %s)" % 
                                         (escape_string(t[1]), 
                                          escape_string(t[3]))))
    elif t[2] == 'like':
        t[0] = qg.Clause(qg.build_clause(t.parser.request, "(tagname = %s AND tagval LIKE %s)" % 
                         (escape_string(t[1]), 
                          escape_string(t[3]))))
    elif t[2] == '~':
        t[0] = qg.Clause(qg.build_clause(t.parser.request, "(tagname = %s AND tagval ~ E%s)" %
                         (escape_string(t[1]), 
                          escape_string(t[3]))))


def p_error(t):
    raise qg.QueryException("Syntax error at '%s'" % t.value, 400)

smapql_parser = yacc.yacc()

class QueryParser:
    """Class to manage parsing and extracting results from the
    database and readingdb"""
    def __init__(self, request): 
        self.parser = copy.copy(smapql_parser)
        self.parser.request = request

    def parse(self, s):
        return self.parser.parse(s, lexer=smapql_lex)

    def runquery(self, db, s, run=True, verbose=False):
        ext, q = self.parse(s)
        if not isinstance(q, list):
            q = [None, q]
            ext = [None, ext]

        if verbose: print q
        if not run: return defer.succeed([])
        
        def eb(error):
            print error
        deferreds = []

        for ext_, q_ in zip(ext[1:], q[1:]):
            if not ext_:
                d = db.runOperation(q_)
                d.addCallback(lambda _: [])
            else:
                d = db.runQuery(q_)
                d.addCallback(ext_)

            d.addErrback(eb)
            deferreds.append(d)

        if len(deferreds) > 1:
            d = defer.DeferredList(deferreds)
            if ext[0]:
                d.addCallback(ext[0])
        else:
            d = deferreds[0]

        return d


if __name__ == '__main__':
    import os
    import readline
    import traceback
    import sys
    import pprint
    import settings as s
    import data
    import atexit
    from twisted.internet import reactor
    from twisted.enterprise import adbapi
    from optparse import OptionParser

    # pull out options
    usage = "usage: %prog [options] archiver-config.ini"
    parser = OptionParser(usage=usage)
    parser.add_option("-n", "--no-action", dest="run", 
                      action="store_false", default=True,
                      help="don't perform queries")
    parser.add_option("-v", "--verbose", dest="verbose", 
                      action="store_true", default=False,
                      help="print generated SQL")
    parser.add_option("-k", "--keys", dest="keys",
                      default="",
                      help="comma-separated list of keys to use in query")
    parser.add_option("-p", "--private", dest="private",
                      default=False, action="store_true",
                      help="request only private results")
    opts, args = parser.parse_args()
    
    if len(args) != 1:
        parser.print_help()
        sys.exit(1)

    s.load(args[0])
    s.import_rdb()

    # set up readline
    HISTFILE = os.path.expanduser('~/.smap-query-history')
    readline.parse_and_bind('tab: complete')
    readline.parse_and_bind('set editing-mode emacs')
    if hasattr(readline, "read_history_file"):
        try:
            readline.read_history_file(HISTFILE)
        except IOError:
            pass
    atexit.register(readline.write_history_file, HISTFILE)

    cp = adbapi.ConnectionPool(s.DB_MOD,
                               host=s.DB_HOST,
                               database=s.DB_DB,
                               user=s.DB_USER,
                               password=s.DB_PASS,
                               cp_min=5, cp_max=15)

    # make a fake request to give the parser with whatever
    # command-line options we need.
    class Request(object):
        pass

    request = Request()
    args = {}
    if opts.keys: args['key'] = opts.keys.split(',')
    if opts.private: args['private'] = []
    setattr(request, 'args', args)

    # make a parser and start reading from the console
    qp = QueryParser(request)

    def readquery():
        try:
            s = raw_input('query > ')   # Use raw_input on Python 2
            if s == '': 
                return readquery()
        except EOFError:
            return reactor.callFromThread(reactor.stop)
        except:
            traceback.print_exc()

        d = qp.runquery(cp, s, verbose=opts.verbose, run=opts.run)
        d.addCallback(lambda v: pprint.pprint(v))
        d.addCallback(lambda x: readquery())
        return d

    d = readquery()
    reactor.run()

    cp.close()



"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

from twisted.internet import reactor
from twisted.web import resource, server
from twisted.web.resource import NoResource
from twisted.python import log

from smap import subscriber, reporting
from smap.server import RootResource
import smap.util as util
from smap.archiver import settings, data, api, republisher


class DataResource(resource.Resource):
    """This resource manages the functionality of the add/ resource,
    which is how data is inserted into the system.  Data is POSTed to
    /add/[api key]; this resource checks that it is a valid key and
    then inserts it into the postgres and readingdb databases.
    """
    def __init__(self, db, republisher):
        self.db = db
        self.republisher = republisher
        self.data = data.SmapData(db)
        resource.Resource.__init__(self)

    def _add_data(self, subid, obj):
        """Add the data using the data interface"""
        if len(obj) == 0: return True
        return self.data.add(subid, obj)

    def _check_subscriber(self, request, subid):
        """Check that there was an API key, and republish"""
        if len(subid) == 1:
            # send the object to the republisher
            public = subid[0][1]
            subid = subid[0][0]
            obj = util.json_decode(request.content.read())
            self.republisher.republish(request.prepath[-1], public, obj)
            reporting.push_metadata(obj)
            return subid, obj
        else:
            request.setResponseCode(404)
            request.finish()
        return server.NOT_DONE_YET

    def getChild(self, name, request):
        return self

    def render_POST(self, request):
        """Handle new data"""

        # first check if the api key is valid
        d = self.db.runQuery("SELECT id, public FROM subscription WHERE key = %s", 
                             (request.prepath[-1],))
        d.addCallback(lambda x: self._check_subscriber(request, x))

        # if so, add the data
        d.addCallback(lambda (subid, obj): self._add_data(subid, obj))
        def add_success(x):
            if not x:
                request.setResponseCode(500)
            request.finish()
        def add_error(x):
            # return a 500 so the sMAP server can hold onto the data
            # until things can be fixed.
            request.setResponseCode(500)
            request.finish()

        # and finish the request
        d.addCallback(add_success)
        d.addErrback(add_error)
        return server.NOT_DONE_YET

def getSite(db):
    """Get the twisted site for smap-archiver"""
    root = RootResource(value={'Contents': ['add', 'api', 'republish']})
    repub = republisher.ReResource(db)
    root.putChild('add', DataResource(db, repub))
    root.putChild('api', api.Api(db))
    root.putChild('republish', repub)
    return server.Site(root)




"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

import sys
import logging
import time
import calendar
import threading
import urllib2
import urlparse
import re
import signal

import smap.driver
from smap.util import periodicSequentialCall

import sensordb

urllib2.install_opener(urllib2.build_opener())

def to_pathname(s):
    s = re.sub('[\W/]+', '_', s)
    s = re.sub('_*$', '', s)
    return s.lower()

class Driver(smap.driver.SmapDriver):
    def setup(self, opts):
        self.url = opts['Url']
        self.auth = (opts['Username'], opts['Password'])
        self.mappings = [x for x in sensordb.DB if x['obviusname'] == opts['ObviusType']][0]
        self.period = opts.get('Period', 30)

        self.set_metadata('/', {
                'Extra/Driver' : 'smap.drivers.obvius',
                'Instrument/Model' : opts['ObviusType']
                })
                          
        # create all the channels
        for (field, regexp, phase, channel,fmt) in self.mappings['sensors'] + self.mappings['meters']:
            self.add_timeseries('/' + phase + '/' + channel, 
                                fmt, data_type='double')
            self.set_metadata('/' + phase, {
                'Extra/Phase' : phase })


    def open_url(self, url, AUTH):
        """Open a URL using urllib2, sending HTTP Basic authentication"""
        mgr = urllib2.HTTPPasswordMgrWithDefaultRealm()
        url_p = urlparse.urlparse(url)
        mgr.add_password(None, url_p.netloc, AUTH[0], AUTH[1])
        handler = urllib2.HTTPBasicAuthHandler(mgr)
        opener = urllib2.build_opener(handler)
        return opener.open(url, timeout=15)

    def is_sensor(self, channel, OBVIUS_SENSORS):
        for (x,y,z,p,q) in OBVIUS_SENSORS:
            if p == channel:
                return True
        return False

    def start(self):
        periodicSequentialCall(self.update).start(self.period)
    
    def update(self):
        logging.debug("Updating " + self.url)

        try:
            fp = self.open_url(self.url, self.auth)
            html = fp.read()
            fp.close()

            # the obvius timestamp is in UTC
            reading_time = re.findall('<time.*?>(.*?)</time>', html)[0]
            # pydoc says this is the right way to convert this timestamp
            reading_time = time.strptime(reading_time + " UTC",
                                         "%Y-%m-%d %H:%M:%S %Z")
            reading_time = calendar.timegm(reading_time)
            logging.debug("reading timestamp is " + str(reading_time))
            
            # this pulls out a list of all the channel-reading pairs
            data = re.findall('<point.*?name="(.*?)".*?value="(.*?)".*?/>', html.lower())
            data_map = {}
            data_map.update(data)

            for (field, regexp, phase, channel,fmt) in \
                    self.mappings['sensors'] + self.mappings['meters']:
                reading = data_map.get(field.lower())
                if not reading: 
                    logging.warn(field + " not found in doc")
                    continue
                match = re.search(regexp, reading)
                if not match: 
                    logging.warn("reading conversion failed: " + reading + " : " + field)
                    continue
                self.add('/' + phase + '/' + channel, reading_time, float(match.groups(0)[0]))

        except IOError, e:
            logging.error("Failed to load meter: %s: %s" % (self.url, str(e)))




"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

import os
import sys
import json
import uuid
from avro import schema, io

import util
import pkgutil

NAMESPACE = "edu.berkeley.cs.local"
SCHEMA_NAMES = schema.Names(default_namespace=NAMESPACE)
SCHEMAS = [
    # new extension types
    "uuid", 
    "UnitofTime", "Duration", 
    "ReadingType", "ReadingValue",
    
    'InstrumentMetadata', 'LocationMetadata',
    
    # timeseries subobjects
    "Actuator", 
    "Properties", "Metadata", 
    "TimeSeries", "Collection",
    "Reporting"]
SCHEMA_OBJECTS = []

# load all the schemas when we check an object
for sf in SCHEMAS:
    # print "Loading", sf
	
    dirs = [os.path.dirname(sys.modules[__name__].__file__), sys.prefix]
    obj = None
    for d in dirs:
        try:
            path = os.path.join(d, 'schema', sf.lower() + '.av')
            obj = json.load(open(path, 'r'))
            break
        except:
            pass
    if obj == None:
        raise Exception('Cannot load schema: ' + sf.lower())
    
    s = schema.make_avsc_object(obj, SCHEMA_NAMES)
    SCHEMA_OBJECTS.append(obj)
        
def validate(schema, obj):
    """Validate an object against its schema.

    Right now, this just checks it against the Avro schema; however in
    the future we will want to impose additional constraints which
    aren't expressable in the schema.
    """
    if schema == 'uuid' and isinstance(obj, uuid.UUID):
        return True
    elif schema == 'Readings':
        return True

    s = SCHEMA_NAMES.get_name(schema, None)
    # swap the uuid for the byte-packed encoding we use with avro
    try:
        id = obj.get('uuid', None)
        if util.is_string(id):
            id = uuid.UUID(id)

        if id: obj['uuid'] = id.bytes
        rv = io.validate(s, obj)
        if id: obj['uuid'] = id
        return rv
    except:
        return False

def filter_fields(schema, obj):
    rv = {}
    schema = util.find(lambda x: x.get("name", "") == schema, SCHEMA_OBJECTS)
    if not schema: 
        return obj

    for field in [x['name'] for x in schema['fields']]:
        if field in obj:
            rv[field] = obj[field]
            
    return rv

# validate("Timeseries", None)



"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

import sys
import urllib2
import urllib
import json
import operator
import pprint
import time
from StringIO import StringIO

import numpy as np
import smap.util as util
import tscache
from smap.core import SmapException

from twisted.internet import reactor
from twisted.internet.protocol import ReconnectingClientFactory
from twisted.internet.endpoints import TCP4ClientEndpoint
from twisted.protocols.basic import LineReceiver
from twisted.web.client import Agent
from twisted.web.http_headers import Headers
from twisted.python import log

try:
    from twisted.web.client import FileBodyProducer
except ImportError:
    from smap.contrib.client import FileBodyProducer

try:
    from smap.iface.http.httpcurl import get
except ImportError:
    print >>sys.stderr, """Warning -- can't find httpcurl
  --> Falling back on urllib2
  --> Install pycURL for parallel data downloads"""
    from smap.iface.http.httputils import get

class SmapClient:
    """Blocking client class for the archiver API.
    """
    def __init__(self, base='http://new.openbms.org/backend', 
                 key=None, private=False, timeout=5.0):
        self.base = base
        self.timeout = timeout
        self.key = key
        self.private = private

    def _build_qdict(self):
        """return a query dict to be passed with all requests"""
        rv = {}
        if self.key:
            rv['key'] = self.key
        if self.private:
            rv['private'] = ''
        return rv

    def query(self, q):
        """Send a query using the ARD query language to the server, and return
the result.

:param str q: the query
:return: the parsed JSON object returned by the server
"""
        try:
            fp = urllib2.urlopen(self.base + '/api/query?' + 
                                 urllib.urlencode(self._build_qdict()),
                                 data=q, 
                                 timeout=self.timeout)
            rv = json.load(fp)
        except urllib2.HTTPError:
            log.err("Bad request running query: ""%s"" " % q)
            raise SmapException()
        fp.close()
        return rv

    def tags(self, qbody, tags='*', nest=False):
        """Look up tags associated with a specific query body"""
        tags = self.query('select %s where %s' % (tags, qbody))
        if not nest:
            return map(lambda t: dict(util.buildkv('', t)), tags)
        else:
            return tags
        
    def _data_uuid(self, uuid, start, end, cache):
        """Construct a list of urls we need to load for a single uuid"""
        if cache:
            cache = tscache.TimeseriesCache(uuid)
            cached_data = cache.read(0, start, end) 
            cache.close()
        else:
            cached_data = []

        cached_data = [((0, start), None)] + \
            cached_data + \
            [((end, 0), None)]
        load_list = []
        for idx in range(0, len(cached_data) - 1):
            fetch_start, fetch_end = cached_data[idx][0][1], cached_data[idx+1][0][0]
            load_list.append([(fetch_start, fetch_end)])
            
        return load_list, map(operator.itemgetter(1), cached_data[1:-1])

    @staticmethod
    def _parser(data):
        obj = util.json_decode(data)
        return obj[0]['uuid'], np.array(obj[0]['Readings'])
        
    def data_uuid(self, uuids, start, end, cache=True):
        """Load a time range of data for a list of uuids
        
Attempts to use cached data and load missing data in parallel.

:param list uuids: a list of stringified UUIDs
:param int start: the timestamp of the first record in seconds, inclusive
:param int end: the timestamp of the last record, exclusive
:return: a list of data vectors.  Each element is
  :py:class:`numpy.array` of data in the same order as the input list of
  uuids
        """
        qdict = self._build_qdict()
        qdict['limit'] = -1
        data, urls = {}, []
        start, end = start * 1000, end * 1000
        now = int((time.time() - 300) * 1000)

        # construct a list of all holes in the cache
        for u in uuids:
            data[u] = self._data_uuid(u, start, end, cache)

            # these are the regions of missing data
            for region in data[u][0]:
                qdict['starttime'] = str(region[0][0])
                qdict['endtime'] = str(region[0][1])
                dlurl = str(self.base + '/api/data/uuid/' + u + '?' +
                            urllib.urlencode(qdict))
                if qdict['starttime'] != qdict['endtime']:
                    region.append(dlurl)
                    urls.append(dlurl)
                else:
                    region.append(None)

        # load all of the missing chunks in parallel
        newdata = dict(get(urls, parser=self._parser, verbose=False))

        # insert all the new data and return the result
        rv = []
        for u in uuids:
            loaddata = []
            for range, url in data[u][0]:
                if url != None and len(newdata[url][1]) > 0:
                    assert newdata[url][0] == u
                    loaddata.append(newdata[url][1])
                    # print "downloaded", len(loaddata[-1])

                    if cache and range[0] < now:
                        c = tscache.TimeseriesCache(u)
                        c.insert(0, range[0], range[1] if range[1] < now else now, 
                                 loaddata[-1][np.nonzero(loaddata[-1][:,0] < now)])
                        c.close()
                else:
                    v = np.array([])
                    v.shape = (0, 2)
                    loaddata.append(v)

            def interleave(x, y):
                lst = [None] * (len(x) * 2 - 1)
                for idx in xrange(0, len(x)):
                    lst[idx * 2] = x[idx]
                for idx in xrange(0, len(y)):
                    lst[idx * 2 + 1] = y[idx]
                return lst
            rv.append(np.vstack(interleave(loaddata, data[u][1])))

        return rv

    def _data(self, selector, where, limit, streamlimit):
        qbody = "select data %s limit %i streamlimit %i where %s" % \
            (selector, limit, streamlimit, where)
        return self.query(qbody)
        

    def data(self, qbody, start, end):
        """Load data for streams matching a particular query

        Uses the local time-series cache in the .cache directory.
        """
        uids = self.query('select distinct uuid where %s' % qbody)
        data = self.data_uuid(uids, start, end)
        return uids, data

    def prev(self, qbody, ref, limit=1, streamlimit=10):
        """Load data before a reference time.

        :param str qbody: a selector identifying the streams to query
        :param int ref: reference timestamp
        :param int limit: the maximum number of points to retrieve per stream
        :param int streamlimit: the maximum number of streams to query
        """
        return self._data("before %i" % (ref * 1000), where, limit, streamlimit)

    def next(self, where, ref, limit=1, streamlimit=10):
        """Load data after a reference time.

        See `prev` for args.
        """
        return self._data("after %i" % (ref * 1000), where, limit, streamlimit)

    def latest(self, where, limit=1, streamlimit=10):
        """Load the last data in a time-series.

        See `prev` for args.
        """
        return self._data("before 4294967295000", where, limit, streamlimit)


class RepublishClient:
    """Listener for streaming data from a sMAP source or archiver's /republish feed
    """
    def __init__(self, url, datacb, reconnect=True, restrict=None, connect_error=None):
        """
:param str url: url of the source
:param datacb: callable to be called with each new sMAP object
:param bool reconnect: weather to reconnect if the socket connection is dropped.
:param str restrict: "where" clause restricting data to be delivered.
    This is only evaluated once, when connecting.
        """
        self.url = url
        self.datacb = datacb
        self.agent = Agent(reactor)
        self.reconnect = reconnect
        self.failcount = 0
        self.restrict = restrict
        self.connect_error = connect_error

    class DataReceiver(LineReceiver):
        """Make our own LineReceiver to read back the streaming data
        from the server.  Use the right delimiter and make sure we can
        handle big objects."""
        MAX_LENGTH = 1e7
        delimiter = '\n\n'

        def __init__(self, client):
            self.client = client

        def lineReceived(self, line):
            self.failcount = 0
            try:
                obj = util.json_decode(line)
                self.client.datacb(obj)
            except:
                log.err()
                print line

        def connectionLost(self, reason):
            self.client._failed()
            self.client._reconnect()

    def _failed(self):
        if self.failcount < 5:
            self.failcount += 1

    def __request(self, response):
        if response.code == 200:
            receiver = RepublishClient.DataReceiver(self)
            receiver.setLineMode()
            response.deliverBody(receiver)
            self.receiver = receiver
        elif callable(self.connect_error):
            self.connect_error(response)

    def _reconnect(self):
        """Exponential backup on the reconnect policy"""
        if self.reconnect and not self.closing:
            print "connection failed, reconnecting in", (self.failcount ** 2)
            reactor.callLater(self.failcount ** 2, self.connect)

    def _connect_failed(self, reason):
        self._failed()
        self._reconnect()

    def connect(self):
        """Subscribe and start receiving data
        """
        self.closing = False
        if not self.restrict:
            d = self.agent.request('GET',
                                   self.url + '/republish',
                                   Headers(),
                                   None)
        else:
            d = self.agent.request('POST',
                                   self.url + '/republish',
                                   Headers(),
                                   FileBodyProducer(StringIO(str(self.restrict))))
        d.addCallback(self.__request)
        d.addErrback(self._connect_failed)
    
    def close(self):
        self.closing = True
        try:
            self.receiver.transport.stopProducing()
        except:
            return False
        return True

# if __name__ == '__main__':
#     def cb(line):
#         print line
#         pass

#     c = RepublishClient('http://smote.cs.berkeley.edu:8079', cb,
#                         restrict="Metadata/SourceName ~ '^410'")
#     c.connect()
#     reactor.run()


if __name__ == '__main__':
    import time
    # c = SmapClient('http://new.openbms.org/backend')
    # c = SmapClient('http://local.cs.berkeley.edu:8079')
    c = SmapClient('http://localhost:8079')
#     print c.tags('Metadata/SourceName = "410 Labjacks"')
#     print c.latest('Metadata/SourceName = "Cory Hall Dent Meters"', limit=1)
    #c._data_uuid('018eba5e-51b6-5a8d-920f-b5a831546610', 
    #             int(time.time()) - 3600 * 24, int(time.time()))
    data = c.data_uuid(['018eba5e-51b6-5a8d-920f-b5a831546610'], 
                       int(time.time()) - 3600 * 300, int(time.time()))
    
    # print data
    print len(np.unique(data[0][:,0])), len(data[0][:,0])



"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

import os
import sys

from zope.interface import implements

from twisted.python import usage
from twisted.plugin import IPlugin
from twisted.application.service import IServiceMaker
from twisted.application import internet
from twisted.internet import reactor, ssl
from twisted.application.service import MultiService

from smap import core, loader, smapconf
from smap.server import getSite

class Options(usage.Options):
    optParameters = [["data-dir", "d", None, "directory for data"],
                     ["port", "p", None, "service port number"],
                     ["sslport", "s", None, "ssl port number"],
                     ["key", "k", None, "ssl server key"],
                     ["cert", "c", None, "ssl crl list"]]

    def parseArgs(self, conf):
        self['conf'] = conf
        if not os.access(self['conf'], os.R_OK):
            print >>sys.stderr, "ERROR: no such configuration file: " + self['conf']
            sys.exit(1)

def getSslContext():
    if smapconf.SERVER["key"] == None or smapconf.SERVER["cert"] == None:
        raise core.SmapException("Cannot create ssl context without key and certificate files")
    
    return ssl.DefaultOpenSSLContextFactory(os.path.expanduser(smapconf.SERVER["key"]), 
                                            os.path.expanduser(smapconf.SERVER["cert"]))

class SmapServiceMaker(object):
    implements(IServiceMaker, IPlugin)
    tapname = "smap"
    description = "A sMAP server"
    options = Options

    def makeService(self, options):
        if options['data-dir'] != None:
            if not os.access(options['data-dir'], os.X_OK | os.W_OK):
                raise core.SmapException("Cannot access " + options['data-dir'])
            smapconf.SERVER['DataDir'] = options['data-dir']

        inst = loader.load(options['conf'])
        # override defaults with command-line args
        smapconf.SERVER.update(dict([(k.lower(), v) for (k, v) in
                                     options.iteritems() if v != None]))

        if 'SuggestThreadPool' in smapconf.SERVER:
            reactor.suggestThreadPoolSize(int(smapconf.SERVER['SuggestThreadPool']))

        inst.start()

        site = getSite(inst)
        service = MultiService()

        # add HTTP and HTTPS servers to the twisted multiservice
        if 'port' in smapconf.SERVER:
            service.addService(internet.TCPServer(int(smapconf.SERVER['port']), site))
        if 'sslport' in smapconf.SERVER:
            service.addService(internet.SSLServer(int(smapconf.SERVER['sslport']), 
                                                  site, 
                                                  getSslContext()))
        return service

serviceMaker = SmapServiceMaker()




import inspect
import numpy as np

import smap.util as util
import smap.operators as operators
import querygen as qg

import smap.operators
import smap.drivers.sumr
import smap.drivers.resamplr

operator_modules = [
    smap.operators,
    smap.drivers.sumr,
    smap.drivers.resamplr,
    ]

installed_ops = {}

def discover():
    for m in operator_modules:
        for name, obj in inspect.getmembers(m):
            if inspect.isclass(obj) and \
                    issubclass(obj, operators.Operator) and \
                    hasattr(obj, "operator_name"):
                installed_ops[obj.operator_name] = obj
    print "found ops:", ', '.join(installed_ops.iterkeys())

def get_operator(name, args):
    """Look up an operator by name.  If given args, try to parse them
    using whatever initializer lists are available.

    :raises: ParseException if it can't match the operator
    """
    groups = filter(lambda x: type(x) == type('') and x[0] == '$', args)
    args = filter(lambda x: type(x) != type('') or x[0] != '$', args)

    if not name in installed_ops:
        raise qg.QueryException("No such operator: " + name)
    if len(args) == 0:
        # if the op doesn't take any args, just return the bare operator
        return installed_ops[name]
    else:
        for proto in installed_ops[name].operator_constructors:
            if len(proto) != len(args): continue
            try:
                alist = map(lambda (fn, a): fn(a), zip(proto, args))
            except ValueError:
                continue
            return lambda inputs: installed_ops[name](inputs, *alist)
        raise qg.QueryException("No valid constructor for operator %s: %s" % 
                                (name, str(args)))

def make_applicator(ops, group=None):
    class _TmpOp(operators.CompositionOperator):
        oplist = ops

    def build_result((d, s)):
        obj = dict(s)
        d[:,0] = np.int_(d[:, 0])
        d[:,0] *= 1000
        obj['Readings'] = d.tolist()
        return util.build_recursive(obj, suppress=[])

    def apply_op(data):
        opmeta = data[0][1]
        opmeta = map(lambda x: dict(util.buildkv('', x)), opmeta)
        if not len(opmeta):
            return []

        # build the operator
        if group and len(group):
            op = operators.GroupByTagOperator(opmeta, _TmpOp, group[0])
        else:
            op = _TmpOp(opmeta)

        # insert the data in the right index for the operator (it
        # could come back in any order)
        opdata = [operators.null] * len(data[1][1])
        for v in data[1][1]:
            if len(v['Readings']):
                idx = op.index(v['uuid'])
                opdata[idx] = np.array(v['Readings'])
                opdata[idx][:, 0] /= 1000

        # process
        redata = op.process(opdata)

        # construct a return value with metadata and data merged
        return map(build_result, zip(redata, op.outputs))

    return apply_op
    
discover()
